#define G_LOG_DOMAIN "ap-ability"

#include "ap-ability.h"

#include <math.h>
#include <glib.h>
#include <glib-object.h>

typedef struct _APAbilityPrivate
{
    gchar *name;
    gint   score;
    gint   bonus;
    gint   modifier;
} APAbilityPrivate;

struct _APAbility
{
    GObject parent_instance;
};

/**
 * SECTION:ap-ability
 * @short_description: Manages a single ability of a character
 * @title: APAbility
 * @include: adventure-pack/ap-ability.h
 *
 * #APAbility is an object for creating and managing a character's ability.
 * Allows the setting of various ability attributes and automatically calculate
 * values like modifiers.
 *
 * # Usage {#usage}
 * ## Basic
 * |[<!-- language="C" -->
 * APAbility *ability;
 *
 * ability = g_object_new (AP_TYPE_ABILITY,
 *                         "name", "Dexterity",
 *                         "score", 10,
 *                         "bonus", 0,
 *                         NULL);
 *
 * g_print ("%s: %i (%i)",
 *          ap_ability_get_name (ability),
 *          ap_ability_get_score (ability),
 *          ap_ability_get_modifier (ability));
 *
 * // Returns: Dexterity: 10 (0)
 * ]|
 *
 * ## Add properties
 * |[<!-- language="C" -->
 * APAbility *ability;
 *
 * ability = ap_ability_new ("dexterity");
 * ap_ability_add_score (ability, 10);
 * ]|
 **/

G_DEFINE_TYPE_WITH_PRIVATE (APAbility, ap_ability, G_TYPE_OBJECT)

enum
{
    PROP_0,
    PROP_NAME,
    PROP_SCORE,
    PROP_BONUS,
    PROP_MODIFER,
    NUM_PROPERTIES
};

static GParamSpec *obj_properties[NUM_PROPERTIES] = { NULL, };

enum
{
    SCORE_CHANGED,
    MODIFER_CHANGED,
    N_SIGNALS
};

static guint signals[N_SIGNALS] = { 0, };

static void
calculate_modifier (APAbility *self)
{
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);
    gint score;
    gint modifier;

    /*
     * When a score number will result in a negative number (ex. 9 - 10 = -1)
     * then we subtract by 1 more to return a proper floor ().
     */
    if (priv->score < 10)
        score = priv->score - 1;
    else
        score = priv->score;

    /*
     * Modifier = ((Ability score) - 10) / 2
     */
    modifier = floor ((score - 10) / 2);

    if (modifier == priv->modifier)
        return;

    priv->modifier = modifier;

    g_signal_emit (self, signals[MODIFER_CHANGED], 0, priv->modifier, priv->name);
}

static void
score_changed (APAbility *self,
               gint       score)
{
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    if (score == priv->score)
        return;

    priv->score = score;

    g_signal_emit (self, signals[SCORE_CHANGED], 0, priv->score, priv->name);

    /* Recalculate modifier */
    calculate_modifier (self);
}

static void
ap_ability_set_property (GObject      *object,
                         guint         property_id,
                         const GValue *value,
                         GParamSpec   *pspec)
{
    APAbility *self = AP_ABILITY (object);
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    switch (property_id)
    {
        case PROP_NAME:
            priv->name = g_value_dup_string (value);
            break;

        case PROP_SCORE:
            score_changed (self, g_value_get_int (value));
            break;

        case PROP_BONUS:
            priv->bonus = g_value_get_int (value);

            //TODO: Propagate score-change when bonus changes
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_ability_get_property (GObject      *object,
                         guint         property_id,
                         GValue       *value,
                         GParamSpec   *pspec)
{
    APAbility *self = AP_ABILITY (object);
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    switch (property_id)
    {
        case PROP_NAME:
            g_value_set_string (value, priv->name);
            break;

        case PROP_SCORE:
            g_value_set_int (value, priv->score + priv->bonus);
            break;

        case PROP_BONUS:
            g_value_set_int (value, priv->bonus);
            break;

        case PROP_MODIFER:
            g_value_set_int (value, priv->modifier);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_ability_finalize (GObject *object)
{
    APAbility *self = AP_ABILITY (object);
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    g_free (priv->name);

    G_OBJECT_CLASS (ap_ability_parent_class)->finalize (object);
}

static void
ap_ability_init (APAbility *self)
{
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    priv->name = NULL;
    priv->score = 0;
    priv->bonus = 0;
    priv->modifier = 0;
}

static void
ap_ability_class_init (APAbilityClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->set_property = ap_ability_set_property;
    object_class->get_property = ap_ability_get_property;
    object_class->finalize = ap_ability_finalize;

	/**
	 * APAbility:name:
	 *
	 * Name of the ability.
     *
     * # Example
     * * `Dexterity`
	 */
    obj_properties[PROP_NAME] =
        g_param_spec_string ("name",
                             "Name of the ability",
                             "Name of the ability.",
                             NULL,
                             G_PARAM_READWRITE);

	/**
	 * APAbility:score:
	 *
	 * Ability score level.
	 */
    obj_properties[PROP_SCORE] =
        g_param_spec_int ("score",
                          "Score of the ability",
                          "Score of the ability.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);

	/**
	 * APAbility:bonus:
	 *
	 * Ability score bonus.
	 */
    obj_properties[PROP_BONUS] =
        g_param_spec_int ("bonus",
                          "Bonus score of the ability",
                          "Bonus score of the ability.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);

	/**
	 * APAbility:modifier:
	 *
	 * (Read-only): Returns the ability modifier
     *
     * modifier = ((#APAbility:score) - 10) / 2
	 */
    obj_properties[PROP_MODIFER] =
        g_param_spec_int ("modifier",
                          "Modifier based of the ability score",
                          "Modifier based of the ability score.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READABLE);

    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       obj_properties);

    /**
     * APAbility::score-changed:
     * @ability: an #APAbility
     * @score: the score of the current @ability
     * @name: the name of the current @ability
     *
     * The ::score-changed signal is emitted each time the @ability<!-- -->'s
     * score has changed.
     */
    signals[SCORE_CHANGED] =
        g_signal_new ("score-changed",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_INT);

    /**
     * APAbility::modifier-changed:
     * @ability: an #APAbility
     * @modifier: the score of the current @ability
     * @name: the name of the current @ability
     *
     * The ::modifier-changed signal is emitted each time the
     * @ability<!-- -->'s modifier has changed.
     */
    signals[MODIFER_CHANGED] =
        g_signal_new ("modifier-changed",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_INT);
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
/**
 * ap_ability_new:
 *
 * Creates a new #APAbility.
 *
 * Returns: (transfer full): an #APAbility instance
 */
APAbility *
ap_ability_new (void)
{
    return g_object_new (AP_TYPE_ABILITY, NULL);
}

/**
 * ap_ability_new_with_name:
 * @name: the ability name for new #APAbility
 *
 * Creates a new #APAbility with @name.
 *
 * Returns: (transfer full): an #APAbility instance
 */
APAbility *
ap_ability_new_with_name (gchar *name)
{
    return g_object_new (AP_TYPE_ABILITY,
                         "name", name,
                         NULL);
}

/**
 * ap_ability_set_name:
 * @self: an #APAbility
 * @name: the ability name for @self
 *
 * Sets the ability name for @self.
 *
 * ## Examples
 * * `Strength`
 * * `Dexterity`
 */
void
ap_ability_set_name (APAbility *self,
                     gchar     *name)
{
    GValue new_name = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));
    g_assert (name != NULL);
    g_assert (g_strcmp0 (name, "") != 0);

    g_value_init (&new_name, G_TYPE_STRING);
    g_value_set_string (&new_name, name);

    g_object_set_property (G_OBJECT (self), "name", &new_name);

    g_value_unset (&new_name);
}

/**
 * ap_ability_get_name:
 * @self: an #APAbility
 *
 * Gets the name of @self.
 *
 * Returns: (transfer full) (nullable): the ability name of @self, free with
 *          g_free()
 */
gchar *
ap_ability_get_name (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "name", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_set_score:
 * @self: an #APAbility
 * @score: the ability score for @self
 *
 * Sets the ability score for @self.
 */
void
ap_ability_set_score (APAbility *self,
                      gint       score)
{
    GValue new_score = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_score, G_TYPE_INT);
    g_value_set_int (&new_score, score);

    g_object_set_property (G_OBJECT (self), "score", &new_score);

    g_value_unset (&new_score);
}

/**
 * ap_ability_get_score:
 * @self: an #APAbility
 *
 * Gets the ability score of @self.
 *
 * Returns: (nullable): the ability score of @self
 */
gint
ap_ability_get_score (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "score", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_set_bonus:
 * @self: an #APAbility
 * @bonus: the ability score bonus for the @self
 *
 * Sets the ability score bonus for @self.
 */
void
ap_ability_set_bonus (APAbility *self,
                      gint       bonus)
{
    GValue new_bonus = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_bonus, G_TYPE_INT);
    g_value_set_int (&new_bonus, bonus);

    g_object_set_property (G_OBJECT (self), "bonus", &new_bonus);

    g_value_unset (&new_bonus);
}

/**
 * ap_ability_get_bonus:
 * @self: an #APAbility
 *
 * Gets the ability score bonus of @self.
 *
 * Returns: (nullable): the ability score bonus of @self
 */
gint
ap_ability_get_bonus (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "bonus", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_get_modifier:
 * @self: an #APAbility
 *
 * Gets the ability score modifier of @self.
 *
 * Returns: (nullable): the ability score modifier of @self
 */
gint
ap_ability_get_modifier (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "modifier", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}
